{
--------------------------------------------------------------------------------
Copyright  Norstedts Juridik AB. All rights reserved.

FILE:
  AccountsDbUnit.pas

REVISION HISTORY:
  050301 KG: Initial version.

PURPOSE OF THIS FILE:

--------------------------------------------------------------------------------
}

unit AccountsDbUnit;

interface

uses
  Windows, Classes, Contnrs, Controls, ComObj, ADODB, Variants,
  AccountsManagerInterfaceUnit, HighestUsedUniqueIdsUnit, AccountsDBCacheUnit,
  AMExtendedTableInfoUnit, AMConstantsUnit, AMSQLUnit, AMNotifierUnit;

type
  TAccountsDatabase = class
  private type
    TWrongAccount = class(TObject)
    // Used with ConvertToVersion1_4
    private class var
      FPeriodLists: IInterfaceList;
    private var
      FAccount: IAccount;
      FFinancialYearUniqueId: Integer;
      FMissingPeriodNumbers: set of Byte;
      class function GetPeriodList(const DB: TAccountsDatabase; FinancialYearUniqueId: Integer): IPeriodList; static;
    public
      constructor Create(const aAccount: IAccount; const aPeriod: IPeriod);
      function Compare(Item2: TWrongAccount): Integer;
      procedure Fix(const DB: TAccountsDatabase);
      procedure AddMissingPeriodNumber(PeriodNumber: Integer);
    end;
  private var
    FGetAccountsManager: TGetAccountsManager;
    FExtendedDbDescriptionFile: string;
    FExtendedTableList: TAMExtendedTableList;
    FDbPath: string;
    FConnection: TADOConnection;
    FCommand: TADOCommand;
    FDataSet: TADODataSet;
    FQuery: TADOQuery; // Not used.
    FStoredProcedures: TObjectList;
    FHighestUsedUniqueIds: THighestUsedUniqueIds;
    FCache: TAccountsDBCache;
    FAMNotifier: TAMNotifier; // Just a reference. The owner is AccountsManager.
    function GetAccountsManager: IAccountsManager;
    function GetCreateTableSQL(const CreateTableSQL: string;
      const TableName: string): string;
    procedure InitExtendedTableList(const ExtendedDbDescriptionFile: string);
    // Not used (and will maybe never be): procedure InsertDefaultValues;
    procedure CreateIndex(const IndexSQL: string);
    procedure CreateSP(const SPSQL: string; Params: array of const);
    procedure DropSP(const SPName: string);
    procedure PrepareParameters(Parameters: TParameters;
      const ParameterValues: array of Variant);
    //function GetUpdateSPSQL(const UpdateSPSQL: string;
    //  const TableName: string): string;
    procedure InitConnection;
    function CompareWrongAccount(Item1, Item2: TObject): Integer;
    procedure ConvertDatabase;
    procedure ConvertToVersion1_3;
    procedure ConvertToVersion1_4;
    procedure ConvertToVersion1_6;
    procedure ConvertToVersion1_7;
    // ConvertToVersion1_8 is not finished yet, but some code is already
    // written, that's temporary outcommented.
    //procedure ConvertToVersion1_8;
    procedure ConvertFromVersion1_8;
    procedure ConvertFromVersion1_9;
    procedure ConvertFromVersion2_0;
  public
    constructor Create(AGetAccountsManager: TGetAccountsManager;
      aAMNotifier: TAMNotifier; const ExtendedDbDescriptionFile: string);
    destructor Destroy; override;
    procedure CreateDatabase(const ADbPath: string);
    function IsOpen: Boolean;
    procedure OpenDatabase(NewDataBase: Boolean; const ADbPath: string = '');
    procedure CloseDatabase;

    function GetDataBaseVersion: string;
    procedure SetDataBaseVersion(const Value: string);

    procedure ValidateExtendedDbDescriptionFile(const NewFilename: string);
    procedure UseStoredExtendedDbDescriptionFile;

    procedure CompactDatabase;
    function GetConnectionString(const ADbPath: string = ''): string;
    function GetNewUniqueId(AUniqueIdName: TUniqueIdName): Integer;
    procedure InitDataSet(const CommandText: WideString;
      const Parameters: array of Variant); overload;
    procedure InitDataSet(const CommandText: WideString); overload;
    procedure ExecuteCommand(const CommandText: WideString;
      const Parameters: array of Variant); overload;
    procedure ExecuteCommand(const CommandText: WideString); overload;
    function InitStoredProc(const ProcName: WideString;
      const Parameters: array of Variant): TADOStoredProc; overload;
    function InitStoredProc(const ProcName: WideString): TADOStoredProc;
      overload;
    procedure ExecuteStoredProc(const ProcName: WideString;
      const Parameters: array of Variant); overload;
    procedure ExecuteStoredProc(const ProcName: WideString); overload;
    procedure ExecuteStoredProc(const ProcName: WideString;
      const Parameters: array of Variant;
      ExtendedTableInfo: IAMExtendedTableInfo); overload;

    procedure NotifyChange(aInterface, aOldInterface: IInterface;
      aAMNCommand: TAMNotifyEvent; InterfaceType: TAMNotifierInterfaceType);
      overload;
    procedure NotifyChange(aInterface: IInterface; aAMNCommand: TAMNotifyEvent;
      InterfaceType: TAMNotifierInterfaceType); overload;

    class function DateToSqlStr(ADateTime: TDateTime): string; static;
    class function BooleanToSqlStr(ABoolean: Boolean): string; static;
    class function ExtendedBooleanToSqlStr(AExtendedBoolean: TExtendedBoolean):
      string; static;
    class function RealToSqlStr(AReal: Currency): string; static;

    property DbPath: string read FDbPath;
    property Cache: TAccountsDBCache read FCache;
    property AccountsManager: IAccountsManager read GetAccountsManager;
    property Connection: TADOConnection read FConnection;
    property Command: TADOCommand read FCommand;
    property DataSet: TADODataSet read FDataSet;
    property Query: TADOQuery read FQuery;
    property ExtendedTableList: TAMExtendedTableList read FExtendedTableList;
  end;

implementation

uses
  SysUtils, StrUtils, DateUtils, StCRC, SortedObjectListUnit,
  // ADO extensions:
  ADODB_TLB, ADOX_TLB, JRO_TLB, DB,
  // Accounts manager package:
  AMMiscellaneousRoutinesUnit, AccountsManagerStringsUnit, AMBalanceUnit,
  AMPeriodUnit;

const
  DatabasePassword = 'justanotherNJpass';
  // A history over the database versions exists for the moment here:
  //   Doc\Miscellaneous info for programmers.pas
  DatabaseVersion = '1.7';

{ TAccountsDatabase }

procedure TAccountsDatabase.CompactDatabase;
var
  CT, TempCT: string;
  TempDbPath: string;
  JROJetEngine: TJROJetEngine;
begin
  CloseDatabase;

  TempDbPath := ExtractFilePath(FDbPath) + '~' + ExtractFileName(FDbPath);
  if FileExists(TempDbPath) then
    DeleteFile(TempDbPath);
  // ndra till Copy_File?
  CopyFile(PChar(FDbPath), PChar(TempDbPath), False);
  if FileExists(TempDbPath) then // (For safety.)
  begin
    DeleteFile(FDbPath);
    CT := GetConnectionString(FDbPath);
    TempCT := GetConnectionString(TempDbPath);
    JROJetEngine := TJROJetEngine.Create(nil);
    try
      JROJetEngine.CompactDatabase(TempCT, CT);
    finally
      JROJetEngine.FreeOnRelease;
      FreeAndNil(JROJetEngine);
    end;
    if FileExists(FDbPath) then
      DeleteFile(TempDbPath)
    else
      // ndra till Copy_File?
      CopyFile(PChar(TempDbPath), PChar(FDbPath), False); // (For safety.)
  end;

  OpenDatabase(False);
end;

function TAccountsDatabase.CompareWrongAccount(Item1, Item2: TObject): Integer;
begin
  Result := (Item1 as TWrongAccount).Compare(Item2 as TWrongAccount);
end;

procedure TAccountsDatabase.ConvertDatabase;
var
  VersionOfDatabaseToConvert: string;

begin
  // A history over the database versions exists for the moment here:
  //   Doc\Miscellaneous info for programmers.pas
  VersionOfDatabaseToConvert := GetDatabaseVersion;

  while VersionOfDatabaseToConvert > DatabaseVersion do
  begin
    if VersionOfDatabaseToConvert = '2.0' then
      ConvertFromVersion2_0
    else
    if VersionOfDatabaseToConvert = '1.9' then
      ConvertFromVersion1_9
    else
    if VersionOfDatabaseToConvert = '1.8' then
      ConvertFromVersion1_8
    else
      break;
    VersionOfDatabaseToConvert := GetDatabaseVersion;
  end;

  // Check that the database isn't to new.
  if VersionOfDatabaseToConvert > DatabaseVersion then
    raise Exception.CreateFmt(FutureDatabaseVersionError,
      [VersionOfDatabaseToConvert]);
  if VersionOfDatabaseToConvert < '1.2' then
  begin
    CreateSP(SP_Query_LedgerEntries_FYear,
      [ProcName_SelectFYearLedgerEntryInfo]);
    CreateSP(SP_Query_LedgerEntries_Period,
      [ProcName_SelectPeriodLedgerEntryInfo]);
  end;

  if VersionOfDatabaseToConvert < '1.3' then
  begin
    ConvertToVersion1_4; // Must run 1.4 conversion before 1.3 in some cases.
    ConvertToVersion1_3;
  end
  else if VersionOfDatabaseToConvert < '1.4' then
    ConvertToVersion1_4;

  if VersionOfDatabaseToConvert < '1.5' then
  begin
    // Recreate SP_Query_LedgerEntries_Period since it has changed, and
    // add SP_Query_LedgerEntries_FYear_Account, SP_Query_LedgerEntries_Period_Account.
    DropSP(SP_Query_LedgerEntries_Period);
    CreateSP(SP_Query_LedgerEntries_Period, [ProcName_SelectPeriodLedgerEntryInfo]);
    CreateSP(SP_Query_LedgerEntries_FYear_Account, [ProcName_SelectFYearLedgerEntryInfoAccount]);
    CreateSP(SP_Query_LedgerEntries_Period_Account, [ProcName_SelectPeriodLedgerEntryInfoAccount]);
  end;

  if VersionOfDatabaseToConvert < '1.6' then
    ConvertToVersion1_6;

  if VersionOfDatabaseToConvert < '1.7' then
    ConvertToVersion1_7;

  (* KG: Not finished and not used yet. This shall soon be done by PA.
  if VersionOfDatabaseToConvert < '1.8' then
    ConvertToVersion1_8;
  *)

  if VersionOfDatabaseToConvert < DatabaseVersion then
    SetDataBaseVersion(DatabaseVersion); // Store the latest database version number.
end;

procedure TAccountsDatabase.ConvertToVersion1_3;
// Before version 1.3, when a financial year balance UnknownLedgerEntries was
// changed, the last period balance UnknownLedgerEntries within the financial
// year was not uppdated, leading to that the sum of the periods
// UnknownLedgerEntries was not the same as the financial years. This method
// compare the sum of period UnknownLedgerEntries within each financial year
// and compare it with the financial years UnknownLedgerEntries. Is there a
// difference, the last period is modified, without modifing the next years
// opening balance.

// Note! The ConvertToVersion1_4 must be run before in order to create
// period balances if they are missing.
var
  AccountQuery : IAccountQuery;
  Accounts     : IAccountList;
  BalanceQuery : IBalanceQuery;
  Balances     : IBalanceList;
  PeriodQuery  : IPeriodQuery;
  I            : Integer;
  J            : Integer;
  K            : Integer;
  Sum          : Currency;
  Balance      : IBalance;
begin
  AccountQuery := AccountsManager.NewAccountQuery;
  AccountQuery.FinancialYearList := AccountsManager.GetFinancialYearList(AccountsManager.NewFinancialYearQuery);
  if AccountQuery.FinancialYearList.Count = 0 then
    Exit;

  Accounts := AccountsManager.GetAccountList(AccountQuery);
  for I := 0 to Pred(Accounts.Count) do
  begin
    BalanceQuery := AccountsManager.NewBalanceQuery;
    BalanceQuery.AccountUniqueId := Accounts[i].UniqueId;
    for J := 0 to Pred(AccountQuery.FinancialYearList.Count) do
    begin
      PeriodQuery := AccountsManager.NewPeriodQuery;
      PeriodQuery.FinancialYearUniqueId := AccountQuery.FinancialYearList[j].UniqueId;
      PeriodQuery.SortOrder := psoOrderNumberWithinFinancialYear;
      BalanceQuery.PeriodList := AccountsManager.GetPeriodList(PeriodQuery);
      Balances := AccountsManager.GetBalanceList(BalanceQuery);
      Sum := 0;
      for K := 0 to Pred(Balances.Count) do
        Sum := Sum + Balances[K].UnknownLedgerEntries;

      Balance := Accounts[i].FinancialYearBalance[PeriodQuery.FinancialYearUniqueId];
      if Sum <> Balance.UnknownLedgerEntries then
        TBalance.GetInstance(Balances[Balances.Count-1]).AddUnknownLedgerEntries(Self, Balance.UnknownLedgerEntries-Sum);
    end;
  end;
end;

procedure TAccountsDatabase.ConvertToVersion1_4;
var
  Query    : IAccountQuery;
  Periods  : IPeriodList;
  Accounts : IAccountList;
  Account  : IAccount;
  I        : Integer;
  J        : Integer;
  WrongAccounts: TSortedObjectList;
  SearchObj: TWrongAccount;
  Index    : Integer;
  Period   : IPeriod;
begin
  Periods := AccountsManager.GetPeriodList(AccountsManager.NewPeriodQuery);
  if Periods.Count = 0 then
    Exit;
  Query := AccountsManager.NewAccountQuery;
  Query.PeriodList := Periods;

  WrongAccounts := TSortedObjectList.Create;
  try
    WrongAccounts.Duplicates := dupIgnore;
    WrongAccounts.CompareEvent := CompareWrongAccount;
    // Fetch all accounts with their period balance.
    Accounts := AccountsManager.GetAccountList(Query);
    // Find all accounts that has missing periods.
    for I := 0 to Pred(Accounts.Count) do
    begin
      Account := Accounts[i];
      for J := 0 to Pred(Periods.Count) do
      begin
        Period := Periods[j];
        if Account.PeriodBalance[Period.UniqueId] = nil then
        begin
          SearchObj := TWrongAccount.Create(Account, Period);
          if WrongAccounts.Search(SearchObj, Index) then
          begin
            SearchObj.Free;
            (WrongAccounts[Index] as TWrongAccount).AddMissingPeriodNumber(Period.PeriodNumber);
          end
          else
            WrongAccounts.Add(SearchObj);
        end;
      end;
    end;

    // Fix the wrong accounts.
    for I := 0 to Pred(WrongAccounts.Count) do
      (WrongAccounts[i] as TWrongAccount).Fix(Self);
  finally
    TWrongAccount.FPeriodLists := nil;
    WrongAccounts.Free;
  end;
end;

procedure TAccountsDatabase.ConvertToVersion1_6;

  // This function also exists in the procedure TJournalEntry.SetJournalEntryId.
  function IdLooksLikeANormalId(const AJournalEntryId: string): Boolean;
  begin
    Result := not (
      AnsiContainsStr(AJournalEntryId, Space) or
      AnsiContainsStr(AJournalEntryId, LParen) or
      AnsiContainsStr(AJournalEntryId, RParen));
  end;

const
  SQL_SelectFromJournalEntries =
    'SELECT UniqueId, JournalEntryId, JournalEntryOriginalId ' +
    'FROM JournalEntries WHERE JournalEntryOriginalId = '''';';
  // Field indecies in the data set.
  JournalEntryIdFI = 1;
  JournalEntryOriginalIdFI = 2;
var
  JournalEntryId: string;
begin
  with FDataSet do
  begin
    CommandType := cmdText;
    CommandText := SQL_SelectFromJournalEntries;
    Open;
    try
      while not Eof do
      begin
        JournalEntryId := Fields[JournalEntryIdFI].AsString;
        if IdLooksLikeANormalId(JournalEntryId) and
          (Length(JournalEntryId) <= FL_JournalEntryOriginalId) then
        begin
          Edit;
          Fields[JournalEntryOriginalIdFI].AsString := JournalEntryId;
          Post;
        end;
        Next;
      end;
    finally
      Close;
    end;
  end;
end;

procedure TAccountsDatabase.ConvertToVersion1_7;
// Removes all but first line in the Miscellaneous database.
begin
  InitDataSet('SELECT UniqueId FROM Miscellaneous;');
  FDataSet.Open;
  try
    FDataSet.Last;
    while FDataSet.RecordCount > 1 do
      FDataSet.Delete;
  finally
    FDataSet.Close;
  end;
end;

(* migrate down from earlier version *)
procedure TAccountsDatabase.ConvertFromVersion1_8;
const
  SQL_AlterTableJournalEntries =
    'ALTER TABLE JournalEntries ' +
      'DROP COLUMN JournalEntrySignature;';
  SQL_AlterTableLedgerEntries1 =
    'ALTER TABLE LedgerEntries ' +
      'DROP COLUMN LedgerEntrySignature;';
  SQL_AlterTableLedgerEntries2 =
    'ALTER TABLE LedgerEntries ' +
      'DROP COLUMN LedgerEntryStatus;';
var
  ParameterNames: string;
  FieldNames: string;
  ValueNames: string;
  UpdateValues: string;
begin
  Assert(GetDataBaseVersion = '1.8');

  // Drop new field to JournalEntries and two to LedgerEntries.
  ExecuteCommand(SQL_AlterTableJournalEntries);
  ExecuteCommand(SQL_AlterTableLedgerEntries1);
  ExecuteCommand(SQL_AlterTableLedgerEntries2);

  // Delete the four changed stored procedures SP_Insert_JournalEntries,
  // SP_Update_JournalEntries, SP_Insert_LedgerEntries and
  // SP_Update_LedgerEntries.
  DropSP(ProcName_InsertIntoJournalEntries);
  DropSP(ProcName_UpdateJournalEntries);
  DropSP(ProcName_InsertIntoLedgerEntries);
  DropSP(ProcName_UpdateLedgerEntries);

  // Recreate stored procedures for JournalEntries.
  FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
    UpdateValues, Table_JournalEntries);
  CreateSP(SP_Insert_JournalEntries, [ProcName_InsertIntoJournalEntries,
    ParameterNames, FieldNames, ValueNames]);
  CreateSP(SP_Update_JournalEntries, [ProcName_UpdateJournalEntries,
    ParameterNames, UpdateValues]);

  // Recreate stored procedures for LedgerEntries.
  FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
    UpdateValues, Table_LedgerEntries);
  CreateSP(SP_Insert_LedgerEntries, [ProcName_InsertIntoLedgerEntries,
    ParameterNames, FieldNames, ValueNames]);
  CreateSP(SP_Update_LedgerEntries, [ProcName_UpdateLedgerEntries,
    ParameterNames, UpdateValues]);

  // downgrade version
  SetDataBaseVersion('1.7');
end;

procedure TAccountsDatabase.ConvertFromVersion1_9;
(*
const
  FYB_Sql = 'SELECT fy.EndDate, fyb.FinancialYearUniqueId, fyb.AccountUniqueId, fyb.OpeningBalance, fyb.LedgerEntriesBalance, fyb.UnknownLedgerEntriesBalance ' +
            'FROM FinancialYears fy INNER JOIN FinancialYearBalances fyb ON fy.UniqueId = fyb.FinancialYearUniqueId ' +
            'ORDER BY FinancialYearUniqueId, AccountUniqueId;';

  PB_Sql =  'SELECT p.FinancialYearUniqueId, p.OrderNumberWithinFinancialYear, p.EndDate, pb.UniqueId, pb.PeriodUniqueId, pb.AccountUniqueId, ' +
            'pb.OpeningBalance, pb.LedgerEntriesBalance, pb.UnknownLedgerEntriesBalance ' +
            'FROM Periods AS p INNER JOIN PeriodBalances AS pb ON p.UniqueId = pb.PeriodUniqueId ' +
            'ORDER BY p.FinancialYearUniqueId, pb.AccountUniqueId, p.OrderNumberWithinFinancialYear;';
var
  FYB_DataSet: TADODataSet;
  PB_DataSet: TADODataSet;

  IB_Period: Currency;
  FirstPeriod: Boolean;
  UpdateAccountPeriods: Boolean;
  FYB_EndDate: TDateTime;
  FYB_FinancialUniqueId: Integer;
  FYB_AccountUniqueId: Integer;
  FYB_OpeningBalance: Currency;
  FYB_LedgerEntriesBalance: Currency;
  FYB_UnknownLedgerEntriesBalance: Currency;
  FYB_UB: Currency;*)
begin
  Assert(GetDataBaseVersion = '1.9');

(* It's not possible to restore old value - so we just do nothing !!!
  FYB_DataSet := TADODataSet.Create(nil);
  FYB_DataSet.Connection := FConnection;
  FYB_DataSet.DisableControls;
  FYB_DataSet.LockType := ltOptimistic;
  FYB_DataSet.CursorType := ctKeyset;
  FYB_DataSet.CommandType := cmdText;
  FYB_DataSet.CommandText := FYB_Sql;

  PB_DataSet := TADODataSet.Create(nil);
  PB_DataSet.Connection := FConnection;
  PB_DataSet.DisableControls;
  PB_DataSet.LockType := ltOptimistic;
  PB_DataSet.CursorType := ctKeyset;
  PB_DataSet.CommandType := cmdText;
  PB_DataSet.CommandText := PB_Sql;

  FYB_DataSet.Open;
  try
    PB_DataSet.Open;
    try
      while not FYB_DataSet.Eof do
      begin
        // FinancialYearUniqueId, AccountUniqueId, OpeningBalance
        FYB_EndDate := FYB_DataSet.Fields[0].AsDateTime;
        FYB_FinancialUniqueId := FYB_DataSet.Fields[1].AsInteger;
        FYB_AccountUniqueId := FYB_DataSet.Fields[2].AsInteger;
        FYB_OpeningBalance := FYB_DataSet.Fields[3].AsCurrency;
        FYB_LedgerEntriesBalance := FYB_DataSet.Fields[4].AsCurrency;
        FYB_UnknownLedgerEntriesBalance := FYB_DataSet.Fields[5].AsCurrency;
        FYB_UB := FYB_OpeningBalance + FYB_LedgerEntriesBalance + FYB_UnknownLedgerEntriesBalance;

        FirstPeriod := True;
        UpdateAccountPeriods := False;
        IB_Period := 0;
        while (not PB_DataSet.Eof) and (FYB_FinancialUniqueId = PB_DataSet.Fields[0].AsInteger) and
              (FYB_AccountUniqueId = PB_DataSet.Fields[5].AsInteger) do
        begin
          if FirstPeriod then
          begin
            UpdateAccountPeriods := FYB_OpeningBalance <> PB_DataSet.Fields[6].AsCurrency;
            IB_Period := FYB_OpeningBalance;
          end;

          if UpdateAccountPeriods then
          begin
            PB_DataSet.Edit;
            PB_DataSet.Fields[6].AsCurrency := IB_Period;
            PB_DataSet.Post;
            // OBS Hantera UnknownLedgers mste sttas om fel p sista perioden
          end;

          IB_Period := IB_Period + PB_DataSet.Fields[7].AsCurrency + PB_DataSet.Fields[8].AsCurrency;
          if FYB_EndDate = PB_DataSet.Fields[2].AsDateTime then
          begin
            // Check if financial years UB is = UB from last period
            if IB_Period <> FYB_UB then
            begin
              // Not same, update last periods unknownladegers to get it same
              PB_DataSet.Edit;
              PB_DataSet.Fields[8].AsCurrency := PB_DataSet.Fields[8].AsCurrency + (FYB_UB - IB_Period);
              PB_DataSet.Post;
            end;
          end;

          FirstPeriod := false;
          PB_DataSet.Next;
        end;
        FYB_DataSet.Next;
      end;
    finally
      PB_DataSet.Close;
      PB_DataSet.Free;
    end;
  finally
    FYB_DataSet.Close;
    FYB_DataSet.Free;
  end;
  *)

  // Downgrade
  SetDataBaseVersion('1.8');
end;

procedure TAccountsDatabase.ConvertFromVersion2_0;
  procedure BatchFromAccountingObjectBalances;
  var
    sSQL: string;
  begin
    // LedgerEntriesBalance could not be restored
    sSQL := 'INSERT INTO AccountingObjectFinancialYearBalances (UniqueId, FinancialYearUniqueId,' +
      'AccountUniqueId, AccountingObjectUniqueId, OpeningBalance, LedgerEntriesBalance, NjArchiveNorstedts) '  +
      'SELECT UniqueId, FinancialYearUniqueId, AccountUniqueId, ' +
      'AccountingObjectUniqueId, OpeningBalance, 0, 0 ' +
      'FROM AccountingObjectBalances';
//    ExecuteCommand(sSQL);
  end;

	procedure BatchFromBalanceTable;
	var
	  sSql: string;
	begin
		sSql := 'INSERT INTO FinancialYearBalances (UniqueId, FinancialYearUniqueId, ' +
		  'AccountUniqueId, OpeningBalance, LedgerEntriesBalance, UnknownLedgerEntriesBalance) ' +
		  'SELECT UniqueId, FinancialYearUniqueId, AccountUniqueId, OpeningBalance, 0, 0'+
		  ' from Balances';
	  ExecuteCommand(sSQL);
	end;


var
  ParameterNames: string;
  FieldNames: string;
  ValueNames: string;
  UpdateValues: string;
begin
  Assert(GetDatabaseVersion = '2.0');

  // 1. Recreate tables

  // AccountingObjectFinancialYearBalances
  FCommand.CommandText := GetCreateTableSQL(
    Table_Create_AccountingObjectFinancialYearBalances,
    Table_AccountingObjectFinancialYearBalances);
  FCommand.Execute;
  // Create SP for AccountingObjectFinancialYearBalances.
  FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
    UpdateValues, Table_AccountingObjectFinancialYearBalances);
  CreateSP(SP_Insert_AccountingObjectFinancialYearBalances,
    [ProcName_InsertIntoAccountingObjectFinancialYearBalances,
    ParameterNames, FieldNames, ValueNames]);
  CreateSP(SP_Update_AccountingObjectFinancialYearBalances,
    [ProcName_UpdateAccountingObjectFinancialYearBalances,
    ParameterNames, UpdateValues]);

  // AccountingObjectPeriodBalances
  FCommand.CommandText := GetCreateTableSQL(
    Table_Create_AccountingObjectPeriodBalances,
    Table_AccountingObjectPeriodBalances);
  FCommand.Execute;
  // Create SP for AccountingObjectPeriodBalances.
  FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
    UpdateValues, Table_AccountingObjectPeriodBalances);
  CreateSP(SP_Insert_AccountingObjectPeriodBalances,
    [ProcName_InsertIntoAccountingObjectPeriodBalances, ParameterNames,
    FieldNames, ValueNames]);
  CreateSP(SP_Update_AccountingObjectPeriodBalances,
    [ProcName_UpdateAccountingObjectPeriodBalances, ParameterNames,
    UpdateValues]);

  // The Balances table corresponds to parts of the
  // TBalanceCollection/TBalance objects. (Not anymore. PA)
  FCommand.CommandText := GetCreateTableSQL(
    Table_Create_FinancialYearBalances, Table_FinancialYearBalances);
  FCommand.Execute;
  // Create SP for FinancialYearBalances.
  FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
    UpdateValues, Table_FinancialYearBalances);
  CreateSP(SP_Insert_FinancialYearBalances,
    [ProcName_InsertIntoFinancialYearBalances, ParameterNames, FieldNames,
    ValueNames]);
  CreateSP(SP_Update_FinancialYearBalances,
    [ProcName_UpdateFinancialYearBalances, ParameterNames, UpdateValues]);

  // PeriodBalances
  FCommand.CommandText := GetCreateTableSQL(
    Table_Create_PeriodBalances, Table_PeriodBalances);
  FCommand.Execute;
  // Create SP for PeriodBalances.
  FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
    UpdateValues, Table_PeriodBalances);
  CreateSP(SP_Insert_PeriodBalances, [ProcName_InsertIntoPeriodBalances,
    ParameterNames, FieldNames, ValueNames]);
  CreateSP(SP_Update_PeriodBalances, [ProcName_UpdatePeriodBalances,
    ParameterNames, UpdateValues]);

  CreateSP(SP_Query_AccountingObjectFinancialYearBalances,
    [ProcName_SelectFromAccountingObjectFinancialYearBalances]);
  CreateSP(SP_Query_AccountingObjectPeriodBalances,
    [ProcName_SelectFromAccountingObjectPeriodBalances]);
  CreateSP(SP_Query_FinancialYearBalances,
    [ProcName_SelectFromFinancialYearBalances]);
  CreateSP(SP_Query_PeriodBalances, [ProcName_SelectFromPeriodBalances]);

  // 3. Batch restore data to new tables
  BatchFromAccountingObjectBalances;
  BatchFromBalanceTable;

  // 4. TODO drop tables and SD.

  SetDataBaseVersion('1.9');
end;

constructor TAccountsDatabase.Create(AGetAccountsManager: TGetAccountsManager;
  aAMNotifier: TAMNotifier; const ExtendedDbDescriptionFile: string);
begin
  inherited Create;
  FGetAccountsManager := AGetAccountsManager;
  FAMNotifier := aAMNotifier;
  FExtendedDbDescriptionFile := ExtendedDbDescriptionFile;
  FDbPath := '';
  FConnection := nil;
  FCommand := nil;
  FDataSet := nil;
  FQuery := nil;
  FHighestUsedUniqueIds := nil;
  FCache := TAccountsDBCache.Create(FGetAccountsManager, Self);
  FStoredProcedures := TObjectList.Create;
  InitExtendedTableList(ExtendedDbDescriptionFile);
end;

procedure TAccountsDatabase.InitConnection;
begin
  // Open the database and create components.
  FConnection := TADOConnection.Create(nil);
  FConnection.ConnectionString := GetConnectionString(FDbPath);
  FConnection.LoginPrompt := False;
  FConnection.Open;
  FCommand := TADOCommand.Create(nil);
  FCommand.Connection := FConnection;
  FDataSet := TADODataSet.Create(nil);
  FDataSet.Connection := FConnection;
  FDataSet.DisableControls;
  FQuery := TADOQuery.Create(nil);
  FQuery.Connection := FConnection;
  FQuery.DisableControls;
end;

procedure TAccountsDatabase.InitDataSet(const CommandText: WideString;
  const Parameters: array of Variant);
begin
  DataSet.CommandType := cmdText;
  if DataSet.CommandText <> CommandText then
    DataSet.CommandText := CommandText;
  PrepareParameters(DataSet.Parameters, Parameters);
end;

procedure TAccountsDatabase.InitDataSet(const CommandText: WideString);
begin
  InitDataSet(CommandText, []);
end;

procedure TAccountsDatabase.InitExtendedTableList(
  const ExtendedDbDescriptionFile: string);
begin
  FExtendedTableList := TAMExtendedTableList.Create;
  if ExtendedDbDescriptionFile <> EmptyStr then
  begin
    if FileExists(ExtendedDbDescriptionFile) then
      FExtendedTableList.ReadTableInfoFromXmlFile(ExtendedDbDescriptionFile)
    else
      FExtendedTableList.ReadTableInfoFromXmlString(ExtendedDbDescriptionFile);
  end;
end;

function TAccountsDatabase.InitStoredProc(const ProcName: WideString;
  const Parameters: array of Variant): TADOStoredProc;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to Pred(FStoredProcedures.Count) do
  begin
    if (FStoredProcedures[I] as TADOStoredProc).ProcedureName = ProcName then
    begin
      Result := FStoredProcedures[I] as TADOStoredProc;
      Break;
    end;
  end;
  if not Assigned(Result) then
  begin
    Result := TADOStoredProc.Create(nil);
    Result.Prepared := False;
    Result.Connection := FConnection;
    Result.ProcedureName := ProcName;
    Result.DisableControls;
    for I := 0 to Pred(Length(Parameters)) do
      Result.Parameters.AddParameter;
    FStoredProcedures.Add(Result);
  end;

  PrepareParameters(Result.Parameters, Parameters);
  if not Result.Prepared then
    Result.Prepared := True;
end;

function TAccountsDatabase.InitStoredProc(const ProcName: WideString):
  TADOStoredProc;
begin
  Result := InitStoredProc(ProcName, []);
end;

(* Not used (and will maybe never be):
// We shall probably not do anything like this.
procedure TAccountsDatabase.InsertDefaultValues;
begin
  // TODO: Hardcoded right now.
  // This should be read from xml file?
  InitStoredProc(ProcName_InsertIntoJournalEntryOrigins,
    [GetNewUniqueId(uinJournalEntryOrigin), 'BK', 'Bokslut']).ExecProc;
  InitStoredProc(ProcName_InsertIntoJournalEntryOrigins,
    [GetNewUniqueId(uinJournalEntryOrigin), 'RE', 'Revision']).ExecProc;
end;
*)

function TAccountsDatabase.IsOpen: Boolean;
begin
  Result := Assigned(FConnection) and (stOpen in FConnection.State);
end;

procedure TAccountsDatabase.NotifyChange(aInterface: IInterface;
  aAMNCommand: TAMNotifyEvent; InterfaceType: TAMNotifierInterfaceType);
begin
  NotifyChange(aInterface, nil, aAMNCommand, InterfaceType);
end;

procedure TAccountsDatabase.NotifyChange(aInterface, aOldInterface: IInterface;
  aAMNCommand: TAMNotifyEvent; InterfaceType: TAMNotifierInterfaceType);
begin
  if not FAMNotifier.IsObserving then
    Exit;
  FAMNotifier.AMNotify(aInterface, aOldInterface, aAMNCommand, InterfaceType);
end;

procedure TAccountsDatabase.OpenDatabase(NewDataBase: Boolean; const ADbPath: string = '');
begin
  CloseDatabase;

  if ADbPath <> EmptyStr then
    FDbPath := ADbPath;

  InitConnection;
  if not NewDataBase then
  begin
    try
      if FExtendedDbDescriptionFile <> '' then
        ValidateExtendedDbDescriptionFile(FExtendedDbDescriptionFile)
      else
        UseStoredExtendedDbDescriptionFile;

      ConvertDatabase; // Make sure we are of latest version.
    except
      CloseDatabase;
      raise;
    end;
  end;
end;

procedure TAccountsDatabase.PrepareParameters(Parameters: TParameters;
  const ParameterValues: array of Variant);
var
  I: Integer;
begin
  for I := 0 to Pred(Length(ParameterValues)) do
  begin
    // To save Null in a column we must specify the column type oterwhise you
    // get an exception.
    // A workaround is to set type as ftVariant in this case.
    if ParameterValues[I] = Null then
      Parameters[I].DataType := ftVariant;
    Parameters[I].Value := ParameterValues[I];
  end;
end;

procedure TAccountsDatabase.CreateDatabase(const ADbPath: string);

  procedure CreateUnknownApplicationRowInJournalEntryOrigins;
  const
    SQL_InsertIntoJournalEntryOrigins =
      'INSERT INTO JournalEntryOrigins ' +
      '(UniqueId, ApplicationId, Description) ' +
      'VALUES (:UniqueId, :ApplicationId, :Description);';
  begin
    // (This is actually unnecessary. It can never occur.)
    if ApplicationIdForUnknownJournalEntryOrigin = EmptyStr then
      raise Exception.Create(
        'TAccountsDatabase.CreateDatabase: ApplicationId must not be empty.');
    // (If journal entry origins are put in the cache in the future, that
    // must be handled here.)
    ExecuteCommand(SQL_InsertIntoJournalEntryOrigins,
      [UniqueIdForUnknownJournalEntryOrigin,
      Copy(ApplicationIdForUnknownJournalEntryOrigin, 1, FL_ApplicationId),
      Copy(DescriptionForUnknownJournalEntryOrigin, 1, FL_Description)]);
  end;

  procedure AddUniqueLineInMiscellaneous;
  // Add the current Database version as the only line in the table
  const
    SQL_InsertIntoMiscellaneous =
      'INSERT INTO Miscellaneous (DatabaseVersion) ' +
      'VALUES (:ADatabaseVersion);';
  begin
    ExecuteCommand(SQL_InsertIntoMiscellaneous,
      [Copy(DatabaseVersion, 1, FL_DatabaseVersion)]);
  end;

  procedure StoreExtendedDbDescriptionFile;
  const
    SQL_SelectFromMiscellaneous =
      'SELECT UniqueId, ExtendedDbDescriptionBlob FROM Miscellaneous;';
  begin
    // (If ExtendedDbDescriptionBlob is put in the cache in the future, that
    // must be handled here.)
    if FileExists(FExtendedDbDescriptionFile) then
    begin
      with FDataSet do
      begin
        CommandType := cmdText;
        CommandText := SQL_SelectFromMiscellaneous;
        Open;
        try
          if not Eof then
          begin
            Edit;
            TMiscRoutines.SaveFileToBlobField(FExtendedDbDescriptionFile,
              FDataSet, 'ExtendedDbDescriptionBlob');
            Post;
          end;
        finally
          Close;
        end;
      end;
    end;
  end;

var
  ADOXCatalog: TADOXCatalog;
  DataSource: string;
  ParameterNames: string;
  FieldNames: string;
  ValueNames: string;
  UpdateValues: string;
  I: Integer;
begin
  try
    CloseDatabase; // Just in case...

    FDbPath := ADbPath;
    if FileExists(FDbPath) then // (Tillflligt.)
      DeleteFile(FDbPath); // (Tillflligt.)

    ForceDirectories(ExtractFilePath(FDbPath));
    // You can read more about DataSource properties in the Accounts manager
    // document "Miscellaneous info for programmers".
    // (Prompt: Maybe not so necessary.)
    DataSource :=
      'Provider=Microsoft.Jet.OLEDB.4.0;' +
      'Data Source=' + FDbPath + ';' +
      'Locale Identifier=' + IntToStr(GetUserDefaultLCID) + ';' +
      'Prompt=' + IntToStr(adPromptNever) + ';' +
      'Jet OLEDB:Engine Type=5;' +
      'Jet OLEDB:Database Password=' + DatabasePassword + ';';
    ADOXCatalog := TADOXCatalog.Create(nil);
    try
      ADOXCatalog.Create1(DataSource);
      ADOXCatalog.Disconnect;
    finally
      ADOXCatalog.Free;
    end;

    InitConnection;
    FConnection.BeginTrans;
    FCommand.CommandType := cmdText;

    //AMExtendedTableList :=
    //  ((FAccountsManager as IGetInstance).GetInstance as
    //  TAccountsManager).GetExtendedTableList;

    // The Miscellaneous table corresponds to parts of the TAccountsModel
    // object (earlier TSieModel object).
    FCommand.CommandText :=
      GetCreateTableSQL(Table_Create_Miscellaneous, Table_Miscellaneous);
    FCommand.Execute;
    AddUniqueLineInMiscellaneous; // Sets the database version
    StoreExtendedDbDescriptionFile;

    // The HighestUsedUniqueIds table corresponds to parts of the
    // TAccountsModel object (earlier TSieModel object).
    THighestUsedUniqueIds.CreateTable(FCommand);

    (* This information shalln't exist in this database.
    // The CompanyInfo table corresponds to the TCompanyInfo object (earlier
    // TSieCompany object).
    FCommand.CommandText :=
      GetCreateTableSQL(Table_Create_CompanyInfo, Table_CompanyInfo);
    FCommand.Execute;
    *)

    (* This information shalln't exist in this database.
    // The FileInfo table corresponds to the TFileInfo object (earlier
    // TSieFileInfo object) plus parts of the TAccountsModel object (earlier
    // TSieModel object).
    FCommand.CommandText :=
      GetCreateTableSQL(Table_Create_FileInfo, Table_FileInfo);
    FCommand.Execute;
    *)

    // The FinancialYears table corresponds to the
    // TFinancialYearCollection/TFinancialYear objects.
    FCommand.CommandText :=
      GetCreateTableSQL(Table_Create_FinancialYears, Table_FinancialYears);
    FCommand.Execute;

    // The Periods table corresponds to the TPeriodsCollection/TPeriod objects.
    FCommand.CommandText :=
      GetCreateTableSQL(Table_Create_Periods, Table_Periods);
    FCommand.Execute;
    CreateIndex(Index_Periods);

    // The Accounts table corresponds to parts of the
    // TAccountCollection/TAccount objects.
    FCommand.CommandText :=
      GetCreateTableSQL(Table_Create_Accounts, Table_Accounts);
    FCommand.Execute;
    CreateIndex(Index_Accounts);

    // The Journals table corresponds to the TJournalCollection/TJournal
    // objects.
    // A possisble field in the future: JournalName.
    // FinancialYearItBelongsTo
    // JournalEntryCollection
    FCommand.CommandText :=
      GetCreateTableSQL(Table_Create_Journals, Table_Journals);
    FCommand.Execute;
    CreateIndex(Index_Journals);

    // Create JournalEntryOrigins table.
    FCommand.CommandText := Table_Create_JournalEntryOrigins;
    FCommand.Execute;
    CreateIndex(Index_JournalEntryOrigins);

    // The JournalEntries table corresponds to the
    // TJournalEntryCollection/TJournalEntry objects.
    FCommand.CommandText :=
      GetCreateTableSQL(Table_Create_JournalEntries, Table_JournalEntries);
    FCommand.Execute;
    CreateIndex(Index_JournalEntries_JournalEntryDate);
    CreateIndex(Index_JournalEntries_JournalEntryId);

    // The LedgerEntries table corresponds to parts of the
    // TLedgerEntryCollection/TLedgerEntry objects.
    FCommand.CommandText :=
      GetCreateTableSQL(Table_Create_LedgerEntries, Table_LedgerEntries);
    FCommand.Execute;
    CreateIndex(Index_LedgerEntries);

    // The AccountingObjectTypes table corresponds to the
    // TAccountingObjectTypeCollection/TAccountingObjectType objects.
    FCommand.CommandText := GetCreateTableSQL(
      Table_Create_AccountingObjectTypes, Table_AccountingObjectTypes);
    FCommand.Execute;
    CreateIndex(Index_AccountingObjectTypes_ParentObjectTypeUniqueId);
    CreateIndex(Index_AccountingObjectTypes_ObjectTypeId);

    // The AccountingObjects table corresponds to the
    // TAccountingObjectCollection/TAccountingObject objects.
    FCommand.CommandText := GetCreateTableSQL(Table_Create_AccountingObjects,
      Table_AccountingObjects);
    FCommand.Execute;
    CreateIndex(Index_AccountingObjects);

    // The Balances table corresponds to parts of the
    // TBalanceCollection/TBalance objects. (Not anymore. PA)
    FCommand.CommandText := GetCreateTableSQL(
      Table_Create_FinancialYearBalances, Table_FinancialYearBalances);
    FCommand.Execute;

    // PeriodBalances
    FCommand.CommandText := GetCreateTableSQL(
      Table_Create_PeriodBalances, Table_PeriodBalances);
    FCommand.Execute;

    // AccountingObjectFinancialYearBalances
    FCommand.CommandText := GetCreateTableSQL(
      Table_Create_AccountingObjectFinancialYearBalances,
      Table_AccountingObjectFinancialYearBalances);
    FCommand.Execute;

    // AccountingObjectPeriodBalances
    FCommand.CommandText := GetCreateTableSQL(
      Table_Create_AccountingObjectPeriodBalances,
      Table_AccountingObjectPeriodBalances);
    FCommand.Execute;

    // The LedgerEntryObjectAssociations table corresponds to the
    // TLedgerEntryObjectAssociationCollection/TLedgerEntryObjectAssociation
    // objects.
    FCommand.CommandText := GetCreateTableSQL(
      Table_Create_LedgerEntryObjectAssociations,
      Table_LedgerEntryObjectAssociations);
    FCommand.Execute;
    CreateIndex(Index_LedgerEntryObjectAssociations);

    // Inte klart. Sjlva tillfllet. En SIE-filsimport t.ex. Kanske inte ska
    // anvndas. Vi ska vl ha ngon sorts ersttare till FileInfo som kanske
    // ska anvndas i stllet.
    (* Not used yet:
    FCommand.CommandText := GetCreateTableSQL(Table_Create_LogOccasions,
      Table_LogOccasions);
    FCommand.Execute;
    *)

    // Inte klart. All info fr att beskriva.
    (* Not used yet:
    FCommand.CommandText :=
      GetCreateTableSQL(Table_Create_LogItems, Table_LogItems);
    FCommand.Execute;
    *)

    // Create SP for Periods table.
    FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
      UpdateValues, Table_Periods);
    CreateSP(SP_Insert_Periods, [ProcName_InsertIntoPeriods, ParameterNames,
      FieldNames, ValueNames]);
    CreateSP(SP_Update_Periods, [ProcName_UpdatePeriods, ParameterNames,
      UpdateValues]);

    // Create SP for FinancialYears table.
    FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
      UpdateValues, Table_FinancialYears);
    CreateSP(SP_Insert_FinancialYears, [ProcName_InsertIntoFinancialYears,
      ParameterNames, FieldNames, ValueNames]);
    CreateSP(SP_Update_FinancialYears, [ProcName_UpdateFinancialYears,
      ParameterNames, UpdateValues]);

    // Create SP for Accounts table.
    FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
      UpdateValues, Table_Accounts);
    CreateSP(SP_Insert_Accounts, [ProcName_InsertIntoAccounts,
      ParameterNames, FieldNames, ValueNames]);
    CreateSP(SP_Update_Accounts, [ProcName_UpdateAccounts,
      ParameterNames, UpdateValues]);

    // Create SP for Journals.
    FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
      UpdateValues, Table_Journals);
    CreateSP(SP_Insert_Journals, [ProcName_InsertIntoJournals,
      ParameterNames, FieldNames, ValueNames]);
    CreateSP(SP_Update_Journals, [ProcName_UpdateJournals,
      ParameterNames, UpdateValues]);

    // Create SP for JournalEntries.
    FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
      UpdateValues, Table_JournalEntries);
    CreateSP(SP_Insert_JournalEntries, [ProcName_InsertIntoJournalEntries,
      ParameterNames, FieldNames, ValueNames]);
    CreateSP(SP_Update_JournalEntries, [ProcName_UpdateJournalEntries,
      ParameterNames, UpdateValues]);

    // Create SP for LedgerEntries.
    FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
      UpdateValues, Table_LedgerEntries);
    CreateSP(SP_Insert_LedgerEntries, [ProcName_InsertIntoLedgerEntries,
      ParameterNames, FieldNames, ValueNames]);
    CreateSP(SP_Update_LedgerEntries, [ProcName_UpdateLedgerEntries,
      ParameterNames, UpdateValues]);

    // Create SP for AccountingObjects.
    FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
      UpdateValues, Table_AccountingObjects);
    CreateSP(SP_Insert_AccountingObjects, [ProcName_InsertIntoAccountingObjects,
      ParameterNames, FieldNames, ValueNames]);
    CreateSP(SP_Update_AccountingObjects, [ProcName_UpdateAccountingObjects,
      ParameterNames, UpdateValues]);

    // Create SP for AccountingObjectTypes.
    FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
      UpdateValues, Table_AccountingObjectTypes);
    CreateSP(SP_Insert_AccountingObjectTypes,
      [ProcName_InsertIntoAccountingObjectTypes, ParameterNames, FieldNames,
      ValueNames]);
    CreateSP(SP_Update_AccountingObjectTypes,
      [ProcName_UpdateAccountingObjectTypes, ParameterNames, UpdateValues]);

    // Create SP for LedgerEntryObjectAssociations.
    FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
      UpdateValues, Table_LedgerEntryObjectAssociations);
    CreateSP(SP_Insert_LedgerEntryObjectAssociations,
      [ProcName_InsertIntoLedgerEntryObjectAssociations, ParameterNames,
      FieldNames, ValueNames]);
    // TODO Ska det inte finnas en
    // ProcName_UpdateIntoLedgerEntryObjectAssociations?

    // Create SP for FinancialYearBalances.
    FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
      UpdateValues, Table_FinancialYearBalances);
    CreateSP(SP_Insert_FinancialYearBalances,
      [ProcName_InsertIntoFinancialYearBalances, ParameterNames, FieldNames,
      ValueNames]);
    CreateSP(SP_Update_FinancialYearBalances,
      [ProcName_UpdateFinancialYearBalances, ParameterNames, UpdateValues]);

    // Create SP for PeriodBalances.
    FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
      UpdateValues, Table_PeriodBalances);
    CreateSP(SP_Insert_PeriodBalances, [ProcName_InsertIntoPeriodBalances,
      ParameterNames, FieldNames, ValueNames]);
    CreateSP(SP_Update_PeriodBalances, [ProcName_UpdatePeriodBalances,
      ParameterNames, UpdateValues]);

    // Create SP for AccountingObjectFinancialYearBalances.
    FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
      UpdateValues, Table_AccountingObjectFinancialYearBalances);
    CreateSP(SP_Insert_AccountingObjectFinancialYearBalances,
      [ProcName_InsertIntoAccountingObjectFinancialYearBalances,
      ParameterNames, FieldNames, ValueNames]);
    CreateSP(SP_Update_AccountingObjectFinancialYearBalances,
      [ProcName_UpdateAccountingObjectFinancialYearBalances,
      ParameterNames, UpdateValues]);

    // Create SP for AccountingObjectPeriodBalances.
    FExtendedTableList.GetExtendedSPSQL(ParameterNames, FieldNames, ValueNames,
      UpdateValues, Table_AccountingObjectPeriodBalances);
    CreateSP(SP_Insert_AccountingObjectPeriodBalances,
      [ProcName_InsertIntoAccountingObjectPeriodBalances, ParameterNames,
      FieldNames, ValueNames]);
    CreateSP(SP_Update_AccountingObjectPeriodBalances,
      [ProcName_UpdateAccountingObjectPeriodBalances, ParameterNames,
      UpdateValues]);

    // Create SP for JournalEntryOrigins. We don't support extended fields here
    // right now.
    CreateSP(SP_Insert_JournalEntryOrigins,
      [ProcName_InsertIntoJournalEntryOrigins]);
    CreateSP(SP_Update_JournalEntryOrigins,
      [ProcName_UpdateJournalEntryOrigins]);

    // Create the row in the JournalEntryOrigins table that stands for
    // "Unknown application".
    CreateUnknownApplicationRowInJournalEntryOrigins;

    // Create some Select queries.
    CreateSP(SP_Query_Periods, [ProcName_SelectUniqueIdFromPeriodsByDate]);

    CreateSP(SP_Query_Journals, [ProcName_SelectUniqueIdFromJournals]);

    CreateSP(SP_Query_FinancialYearBalances,
      [ProcName_SelectFromFinancialYearBalances]);

    CreateSP(SP_Query_PeriodBalances, [ProcName_SelectFromPeriodBalances]);

    CreateSP(SP_Query_AccountingObjectFinancialYearBalances,
      [ProcName_SelectFromAccountingObjectFinancialYearBalances]);

    CreateSP(SP_Query_AccountingObjectPeriodBalances,
      [ProcName_SelectFromAccountingObjectPeriodBalances]);

    CreateSP(SP_Query_LedgerEntries_FYear,
      [ProcName_SelectFYearLedgerEntryInfo]);

    CreateSP(SP_Query_LedgerEntries_FYear_Account,
      [ProcName_SelectFYearLedgerEntryInfoAccount]);

    CreateSP(SP_Query_LedgerEntries_Period,
      [ProcName_SelectPeriodLedgerEntryInfo]);

    CreateSP(SP_Query_LedgerEntries_Period_Account,
      [ProcName_SelectPeriodLedgerEntryInfoAccount]);

    // Now add all custom indexes from the XML file.
    for I:= 0 to Pred(FExtendedTableList.IndexListSQL.Count) do
    begin
      if FExtendedTableList.IndexListSQL[I] <> EmptyStr then
        CreateIndex(FExtendedTableList.IndexListSQL[I]);
    end;

    FConnection.CommitTrans;

{
    OpenDatabase(ADbPath);
    FCommand.CommandText :=
      'INSERT INTO FinancialYears (UniqueId, BeginDate, EndDate) ' +
      'VALUES (' +
        IntToStr(1) + ', ' +
        DateToSqlStr(StrToDate('2004-01-01')) + ', ' +
        DateToSqlStr(StrToDate('2004-12-31')) + ')';
    FCommand.Execute;
    CloseDatabase;
}

//    OpenDatabase(ADbPath);
    THighestUsedUniqueIds.InitTable(FCommand);

    // Not used (and will maybe never be): InsertDefaultValues;

    CloseDatabase;
  except
    on E: Exception do
      raise Exception.Create('Error in TAccountsDatabase.CreateDatabase. ' +
        'Error message:' + CRLF + E.Message);
  end;
end;

procedure TAccountsDatabase.CreateIndex(const IndexSQL: string);
begin
  try
    FCommand.CommandText := IndexSQL;
    FCommand.Execute;
  except // We get this exception if index already exist but that is ok.
    // Errorcode -2147217865.
    on E: EOleException do ;
  else
    raise;
  end;
end;

procedure TAccountsDatabase.CreateSP(const SPSQL: string;
  Params: array of const);
begin
  DropSP(AnsiString(Params[0].VAnsiString));
  FCommand.CommandText := Format(SPSQL, Params);
  FCommand.Execute;
end;

destructor TAccountsDatabase.Destroy;
begin
  FreeAndNil(FCache);
  CloseDatabase;
  FStoredProcedures.Free;
  if Assigned(FExtendedTableList) then
    FExtendedTableList.Free;
  inherited;
end;

procedure TAccountsDatabase.DropSP(const SPName: string);
begin
  try
    FCommand.CommandText := 'DROP PROCEDURE ' + SPName;
    FCommand.Execute;
  except // We get this exception if query does not exist, which is ok.
    // If we should suppress all exceptions we can remove ComObj from uses list.
    // Errorcode -2147217865.
    on E: EOleException do ;
  else
    raise;
  end;
end;

class function TAccountsDatabase.BooleanToSqlStr(ABoolean: Boolean): string;
begin
  if ABoolean then
    Result := 'True'
  else
    Result := 'False';
end;

procedure TAccountsDatabase.CloseDatabase;
var
  I: Integer;
begin
  if Assigned(FCommand) and Assigned(FHighestUsedUniqueIds) then
    FHighestUsedUniqueIds.Save(FCommand);
  FreeAndNil(FHighestUsedUniqueIds);

  if Assigned(FCommand) then
    FreeAndNil(FCommand);
  if Assigned(FDataSet) then
  begin
    FDataSet.Close;
    FreeAndNil(FDataSet);
  end;
  if Assigned(FQuery) then
  begin
    FQuery.Close;
    FreeAndNil(FQuery);
  end;
  if Assigned(FStoredProcedures) then
  begin
    for I := 0 to Pred(FStoredProcedures.Count) do
      (FStoredProcedures[I] as TADOStoredProc).Close;
    FStoredProcedures.Clear;
  end;
  if Assigned(FConnection) then
  begin
    FConnection.Close;
    FreeAndNil(FConnection);
  end;
end;

function TAccountsDatabase.GetAccountsManager: IAccountsManager;
begin
  Result := FGetAccountsManager as IAccountsManager;
end;

function TAccountsDatabase.GetCreateTableSQL(const CreateTableSQL: string;
  const TableName: string): string;
var
  ExtendedTableSQL: string;
begin
  ExtendedTableSQL :=
    FExtendedTableList.GetExtendedFieldCreateTableSQL(TableName);
  Result := CreateTableSQL + ExtendedTableSQL + ')';
end;

function TAccountsDatabase.GetDataBaseVersion: string;
begin
  InitDataSet('SELECT UniqueId, DatabaseVersion FROM Miscellaneous;');
  FDataSet.Open;
  try
    FDataSet.Last; // Fix for bug previous version 1.7
    Result := FDataSet.FieldByName('DatabaseVersion').AsString
  finally
    FDataSet.Close;
  end;
end;

function TAccountsDatabase.GetConnectionString(const ADbPath: string = ''):
  string;
begin
  // You can read more about DataSource properties in the Accounts manager
  // document "Miscellaneous info for programmers".
  Result :=
    'Provider=Microsoft.Jet.OLEDB.4.0;' +
    'Data Source=' + IfThen(ADbPath <> EmptyStr, ADbPath, FDbPath) + ';' +
    'Locale Identifier=' + IntToStr(GetUserDefaultLCID) + ';' +
    'Prompt=' + IntToStr(adPromptNever) + ';' +
    'Jet OLEDB:Engine Type=5;' +
    'Jet OLEDB:Database Password=' + DatabasePassword + ';';
end;

function TAccountsDatabase.GetNewUniqueId(AUniqueIdName: TUniqueIdName):
  Integer;
begin
  if not Assigned(FHighestUsedUniqueIds) then
    FHighestUsedUniqueIds := THighestUsedUniqueIds.Create(FDataSet);
  Result := FHighestUsedUniqueIds.GetNewUniqueId(AUniqueIdName);
end;

class function TAccountsDatabase.DateToSqlStr(ADateTime: TDateTime): string;
const
  // (Some characters have special meanings when Delphi's DateTimeToString
  // function is used. They must therefore be enclosed by quotes to not have
  // those meanings. "/" and ":" have special meanings today. "#" has no
  // special meaning today, but has despit that here been enclosed just for
  // safety, to handle the case that it maybe in a future Delphi version will
  // have a special meaning.)
  DateFormatInSQLStatements = '"#"m"/"d"/"yyyy"#"';
  DateTimeFormatInSQLStatements = '"#"m"/"d"/"yyyy h":"n":"s"#"';
begin
  if TimeOf(ADateTime) = 0 then
    DateTimeToString(Result, DateFormatInSQLStatements, ADateTime)
  else
    DateTimeToString(Result, DateTimeFormatInSQLStatements, ADateTime);
end;

class function TAccountsDatabase.ExtendedBooleanToSqlStr(
  AExtendedBoolean: TExtendedBoolean): string;
begin
  case AExtendedBoolean of
    ebFalse:
      Result := 'False';
    ebTrue:
      Result := 'True';
  else // (ebBoth)
    raise Exception.Create('TAccountsDatabase.ExtendedBooleanToSqlStr: ' +
      'Incorrect parameter.');
  end;
end;

class function TAccountsDatabase.RealToSqlStr(AReal: Currency): string;
begin
  Str(AReal, Result); // Period as decimal separator, no thousand separator.
end;

procedure TAccountsDatabase.SetDataBaseVersion(const Value: string);
begin
  InitDataSet('SELECT UniqueId, DatabaseVersion FROM Miscellaneous;');
  FDataSet.Open;
  try
    FDataSet.Edit;
    FDataSet.FieldByName('DatabaseVersion').AsString := Value;
    FDataSet.Post;
  finally
    FDataSet.Close;
  end;
end;

procedure TAccountsDatabase.UseStoredExtendedDbDescriptionFile;
const
  SQL_SelectFromMiscellaneous =
     'SELECT UniqueId, ExtendedDbDescriptionBlob FROM Miscellaneous;';
var
  Stream: TStream;
begin
  // (If ExtendedDbDescriptionBlob is put in the cache in the future, that
  // must be handled here.)
  with FDataSet do
  begin
    CommandType := cmdText;
    CommandText := SQL_SelectFromMiscellaneous;
    Open;
    try
      if not Eof and not FieldByName('ExtendedDbDescriptionBlob').IsNull then
      begin
        Stream := DataSet.CreateBlobStream(
          DataSet.FieldByName('ExtendedDbDescriptionBlob'), bmRead);
        try
          Stream.Position := 0;
          FExtendedTableList.ReadTableInfoFromStream(Stream);
        finally
          Stream.Free;
        end;
      end;
    finally
      Close;
    end;
  end;
end;

procedure TAccountsDatabase.ValidateExtendedDbDescriptionFile(
  const NewFilename: string);
const
  SQL_SelectFromMiscellaneous =
    'SELECT UniqueId, ExtendedDbDescriptionBlob FROM Miscellaneous;';
var
  CRCForNewFile: Longint;
  CRCForOldFile: Longint;
begin
  // (If ExtendedDbDescriptionBlob is put in the cache in the future, that
  // must be handled here.)
  with DataSet do
  begin
    CommandType := cmdText;
    CommandText := SQL_SelectFromMiscellaneous;
    Open;
    try
      if not Eof and not FieldByName('ExtendedDbDescriptionBlob').IsNull then
        CRCForOldFile := TMiscRoutines.CRCForFileInBlobField(
          DataSet, 'ExtendedDbDescriptionBlob')
      else
        Exit;
    finally
      Close;
    end;
  end;
  CRCForNewFile := StCRC.Crc32OfFile(NewFileName);
  if CRCForNewFile <> CRCForOldFile then
    raise Exception.Create('The extended database description file differ ' +
      'from the earlier used file.');
end;

procedure TAccountsDatabase.ExecuteCommand(const CommandText: WideString;
  const Parameters: array of Variant);
begin
  FCommand.CommandType := cmdText;
  if FCommand.CommandText <> CommandText then
    FCommand.CommandText := CommandText;
  PrepareParameters(FCommand.Parameters, Parameters);
  FCommand.Execute;
end;

procedure TAccountsDatabase.ExecuteCommand(const CommandText: WideString);
begin
  ExecuteCommand(CommandText, []);
end;

procedure TAccountsDatabase.ExecuteStoredProc(const ProcName: WideString;
  const Parameters: array of Variant);
begin
  InitStoredProc(ProcName, Parameters).ExecProc;
end;

procedure TAccountsDatabase.ExecuteStoredProc(const ProcName: WideString);
begin
  ExecuteStoredProc(ProcName, []);
end;

procedure TAccountsDatabase.ExecuteStoredProc(const ProcName: WideString;
  const Parameters: array of Variant; ExtendedTableInfo: IAMExtendedTableInfo);
var
  I: Integer;
  ArrayLength: Integer;
  ExtendedParameter: array of Variant;
begin
  if ExtendedTableInfo.Count <> 0 then
  begin
    ArrayLength := Length(Parameters);
    SetLength(ExtendedParameter, ArrayLength + ExtendedTableInfo.Count);
    for I := 0 to Pred(Length(Parameters)) do
      ExtendedParameter[I] := Parameters[I];

    for I := 0 to Pred(ExtendedTableInfo.Count) do
      ExtendedParameter[ArrayLength + I] :=
        ExtendedTableInfo.ExtendedFieldInfo[I].Value;
    InitStoredProc(ProcName, ExtendedParameter).ExecProc;
  end
  else
    InitStoredProc(ProcName, Parameters).ExecProc;
end;

{ TAccountsDatabase.TWrongAccount }

procedure TAccountsDatabase.TWrongAccount.AddMissingPeriodNumber(PeriodNumber: Integer);
begin
  FMissingPeriodNumbers := FMissingPeriodNumbers + [PeriodNumber]
end;

function TAccountsDatabase.TWrongAccount.Compare(Item2: TWrongAccount): Integer;
begin
  Result := AnsiCompareStr(FAccount.AccountId, Item2.FAccount.AccountId);
  if Result = 0 then
    Result := FFinancialYearUniqueId-Item2.FFinancialYearUniqueId;
end;

constructor TAccountsDatabase.TWrongAccount.Create(const aAccount: IAccount; const aPeriod: IPeriod);
begin
  inherited Create;
  FAccount := aAccount;
  FFinancialYearUniqueId := aPeriod.FinancialYearUniqueId;
  AddMissingPeriodNumber(aPeriod.PeriodNumber);
end;

procedure TAccountsDatabase.TWrongAccount.Fix(const DB: TAccountsDatabase);
var
  PeriodList             : IPeriodList;
  ExistingsPeriodNumbers : set of byte;
  I                      : Integer;
begin
  ExistingsPeriodNumbers := [];
  PeriodList := GetPeriodList(DB, FFinancialYearUniqueId);
  for I := 0 to pred(PeriodList.Count) do
  begin
    if not (PeriodList[i].PeriodNumber in FMissingPeriodNumbers) then
      Include(ExistingsPeriodNumbers, PeriodList[i].PeriodNumber);
  end;
  TBalance.AddBalancesForAccount(DB, FAccount, PeriodList, ExistingsPeriodNumbers);
end;

class function TAccountsDatabase.TWrongAccount.GetPeriodList(const DB: TAccountsDatabase;
  FinancialYearUniqueId: Integer): IPeriodList;
// Handles a simple cache.
var
  Query : IPeriodQuery;
  I     : Integer;
begin
  if FPeriodLists = nil then
    FPeriodLists := TInterfaceList.Create;
  for I := 0 to Pred(FPeriodLists.Count) do
  begin
    if (FPeriodLists[i] as IPeriodList)[0].FinancialYearUniqueId = FinancialYearUniqueId then
    begin
      Result := FPeriodLists[i] as IPeriodList;
      Exit;
    end;
  end;
  Query := DB.AccountsManager.NewPeriodQuery;
  Query.FinancialYearUniqueId := FinancialYearUniqueId;
  Result := DB.AccountsManager.GetPeriodList(Query);
  FPeriodLists.Add(Result);
end;

end.
